/* apmakemarket2.cpp
   make a market (if possible) for tradesecurity
*/

/* 
   algorithm:
     this makes a market in one security (tradesecurity)
     initially, each investor's buy price and sell price is set equal to the reservation price, based on the ratio of the
        marginal utility for a unit of the security divided by the marginal utility of consumption
     the marginal utility of the security is a weighted average of the state marginal utilities, using the security payoffs
        as weights
     constraints on minimum consumptions are used to set buy prices to -infinity if no units can be bought and to set
        sell prices to +infinity if no units can be sold
     for trade to be possible, the max buy price must exceed the min sell price by more than the max price spread
         note: the use of infinity for infeasible buys and sells restricts the range to feasible transactions
     if no trade is possible, no trades are made
     
     if trades are possible, aptrade2() is called to determine the maximum amount each investor is willing and able to buy or sell
        these are signed quantities (+ for buy, - for sell) in units of the tradesecurity
        the netamount returned by aptrade2 is the net demand (a signed quantity)t equals quantity sold
        
     after the trades are made, the consumptions matrix is updated   
*/

void apmakemarket2()
{
   int inv,state,sec,consrow,secrow,portrow;
   double v,maxpbuy,minpsell,phigh,plow,nethigh,netlow,pnext,netnext;
   double demand,supply,propdemand,propsupply,prop,amt,tradenet;
   char c;
   double sumb,sums,meanb,means,netamount,aa,bb,pbest,bestamount;
   int numb,nums,done,ok;
   double mincons;
   int debug,ntimes;
   debug = 0;
   
  // for each investor set buy and sell prices and adjust to infinity or -infinity if minconsumptions would be violated
        for (inv=0;inv<ninvestors;inv++)
        {
          // set rows
            secrow = nsecurities*inv;
            consrow = nstates*inv;
          // set buy and selll prices to reservation price
             v = respsecurities[secrow+tradesecurity];
             #ifdef MEMCHECK
                 if (inv>buypsize) aperrorhalt("Memory error: apmakemarket2 01");
             #endif   
             buyp[inv] = v;
             #ifdef MEMCHECK
                 if (inv>sellpsize) aperrorhalt("Memory error: apmakemarket2 02");
             #endif   
             sellp[inv] = v;
         // if current consumption is at minimum, no purchase is feasible, set buyp to -infinity (no purchase at any price)
             if (consumptions[consrow+0]<=minconsumptions[consrow+0]) buyp[inv] = -INFINITY;
          // if any future consumption is at minimum, no sale is feasible if security has a positive payoff in that state
             for (state=0;state<nstates;state++)
               if ((consumptions[consrow+state]<=minconsumptions[consrow+state])&&(securities[nsecurities*state+tradesecurity]>0))
                  sellp[inv] = INFINITY;
        }        
  
  // find maximum buy price and minimum sell price
       maxpbuy  = buyp[0];
       minpsell = sellp[0];
       for (inv=1;inv<ninvestors;inv++)
       { 
          if (buyp[inv]>maxpbuy)   maxpbuy  = buyp[inv]; 
          if (sellp[inv]<minpsell) minpsell = sellp[inv];
        }
   
   // set initial high and low prices
       phigh = maxpbuy;
       plow = minpsell;
       
  // if no trades are feasible, terminate 
       if (phigh <= plow)
       {
          tradeamount = 0;
          return;
       }
       
 // compute mean of buy prices in range and mean of sell prices in range from minpsell to maxpbuy
       sumb = 0; numb = 0; sums = 0; nums = 0;
       for (inv=0;inv<ninvestors;inv++)
       {  if ((buyp[inv]<=maxpbuy)&&(buyp[inv]>=minpsell))    { sumb = sumb+buyp[inv]; numb = numb+1;}
          if ((sellp[inv]>=minpsell)&&(sellp[inv]<=minpsell)) { sums = sums+sellp[inv]; nums=nums+1; }
       }
       meanb = sumb/numb;
       means = sums/nums;
  
  // iterate until done
     done = 0;
     ntimes = 0;
     bestamount = -1;
     while (done == 0)
     {
       // compute next price as average of mean of buy and mean of sell prices in range if first time through
          if (ntimes == 0) pnext = (meanb + means) / 2;
       // if not first time through, compute next price as average of high and low
          if (ntimes >= 1) pnext = 0.5*(phigh + plow);
          
      // find net amount at pnext
          tradeprice = pnext;
          netnext = aptrade2();
      // save if best thus far
          if (tradeamount > bestamount)
          {
             pbest = tradeprice;
             bestamount  = tradeamount;
          }
          
      // increment ntimes
          ntimes = ntimes + 1;
      // set new high and low prices
         // if net amount is positive, make this price plow      
            if (netnext > 0)
            {
              plow = pnext;
              netlow = netnext;
            }
         // if net amount is negative, make this price phigh   
            if (netnext <= 0)
            {
              phigh = pnext;
              nethigh = netnext;
            }

      // conditions for termination
        // if net amount = 0 for tradeprice it is the optimal price
            if (netnext == 0) done = 1;
        // if new range is smaller than maxpricespread, set done to 1
            if ((phigh - plow) < MAXPRICESPREAD * 0.5 *(phigh + plow)) done = 1;
        // if max number of prices in price discovery is reached, set done to 1
            if (ntimes >= NUMPRICES) done = 1;

       // if done, set tradeprice to best and recalculate if necessary
           if (pbest != tradeprice)
           {
             tradeprice = pbest;
             netnext = aptrade2();
           }
         
   } // while (done == 0)   
   
     // set security price to tradeprice
         #ifdef MEMCHECK
              if (tradesecurity>securitypricessize) aperrorhalt("Memory error: apmakemarket2 03");
         #endif   
         securityprices[tradesecurity] = tradeprice;

     // find proportions traded   
         propdemand = 0;
         if (totaldemand > 0.0) propdemand = tradeamount/totaldemand;
         propsupply = 0;
         if (totalsupply > 0.0) propsupply = tradeamount/totalsupply;
         
     // make trades
          for (inv=0;inv<ninvestors;inv++)
          {
             portrow = nsecurities*inv;
             prop = propdemand;
             if (tradev[inv]<0) prop=propsupply;
             amt = prop*tradev[inv];
             #ifdef MEMCHECK
                 if (portrow+tradesecurity>portfoliossize) aperrorhalt("Memory error: apmakemarket2 04");
             #endif   
             portfolios[portrow+tradesecurity] = portfolios[portrow+tradesecurity] + amt;
             portfolios[portrow+0]             = portfolios[portrow+0]             - amt*tradeprice;
          }   
             
     // compute new consumption matrix
             apgetcons();
             
      // compute new security and state reservation prices       
             apgetrps();
 
 }