/* aptrade2.cpp
   computes tradev, a vector of the amounts each investor is willing and able to trade of security tradesecurity at tradeprice
   netamount -- the net demand (demand - supply)
   totaldemand -- the total quantity demanded
   totalsupply -- the total quantity supplied (non negative)
   tradeamount -- the smaller of totaldemand and totalsupply
*/

/*
   algorithm:
     given a tradeprice, threshhold prices are determined
        the buy price threshhold is the tradeprice times (1 + .5*maxpricerange);
        the sell price threshhold is the tradeprice times (1 - .5*maxpricerange);
     if an investor's buy price is below the buy threshhold, the amount bought = 0
     if an investor's sell price is above the sell threshhold, the amount sold = 0
     if an investor will neither sell nor buy anything, his or her amount traded = 0
     if an investor will sell or buy, the amount is determined
     The general principles follow from the fact that all marginal utility functions are monotonic and continuous
     given a tradesecurity and tradeprice, the rate of change of expected marginal utility per unit change in the holding
        of tradesecurity can be computed, this is termed newmu
     the value of newmu can be computed for any amount of the security traded; the amount traded is termed k
        positive values of k represent purchases, negative values are sales
     each investor is processed in turn, to determine the maximum absolute value he or she is willing and able to trade
     if an investor's buy price is above the tradeprice, he or she will buy (k>0)
     if an investor's sell price is below the tradeprice, he or she will sell (k<0)
     an investor will be either a buyer, a seller, or neither
        if neither, the amount traded is set to zero
     if an investor is going to trade, the initial computation sets the minimum and maximum possible k values (kmin and kmax)
        for a buyer, kmin=0 and kmax is positive
        for a seller, kmax=0 and kmin is negative
     the non-zero end of each range is computed so that minimum consumption requirements are not violated
        for a buyer, the value of kmax would cause current consumption to equal the minimum allowed
        for a seller, the value of kmin would cause consumption in some future state to equal the minimum allowed
     for a buyer, the goal is to purchase an amount that will make the reservation price of the security being traded more 
        than the trade price and less than the trade price plus half the maximum percentage price range
        If this is not possible since the reservation price is above the threshhold for all feasible amounts, the amount with
        a reservation price closest to the trade price is used
     for a seller, the goal is to sell an amount that will make the reservation price of the security being traded less 
        than the trade price and more than the trade price minus half the maximum percentage price range
        If this is not possible since the reservation price is below the threshhold for all feasible amounts, the amount with
        a reservation price closest to the trade price is used
     For both buyers and sellers, a simple algorithm is used in which the feasible trading range is bisected at each iteration,
        with either the left or right point replaced so that one reservation price is above the threshhold and another is below it
     The process stops when the smaller amount traded gives a reservation price within the desired range.
*/

double aptrade2()
{
   int inv,state,security,j,done,consrow;
   double kleft,kright,kmid;
   double a,lambda,kmin,kmax,sval,kval;
   double rpleft,rpright,rpmid;
   double buythresh,sellthresh;
   double aa,bb,netamount;
   char c;
   double consx,mincons; 
   int skip;
   
  // initialize net amount, total demand and total supply
     netamount = 0;
     totaldemand = 0;
     totalsupply = 0;
     
  // compute threshholds for buy price and sell price
     buythresh =  tradeprice * (1 + 0.5*MAXPRICESPREAD);
     sellthresh = tradeprice * (1 - 0.5*MAXPRICESPREAD);
  
  // create vector of changes in consumption from gettting 1 unit of tradesecurity and giving up tradeprice units of security 0  
     for (state=0;state<nstates;state++)
     {
       #ifdef MEMCHECK
         if (state>conschgsize) aperrorhalt("Memory error: aptraded2 01");
       #endif   
       conschg[state] = securities[nsecurities*state+tradesecurity] - tradeprice*securities[nsecurities*state+0];
     }  

  // find amounts investors will trade        
    for (inv=0;inv<ninvestors;inv++)
    {
         // set skip condition
             skip = 0;
         //set row
            consrow = nstates*inv;
            
         // set amount traded to zero
             #ifdef MEMCHECK
                if (inv>tradevsize) aperrorhalt("Memory error: aptraded2 02");
             #endif   
             tradev[inv] = 0;

        // see if investor will trade -- if not, leave tradev at zero and go to next investor
             if ((buyp[inv] < buythresh) && (sellp[inv] > sellthresh)) skip = 1;
             
      if (skip ==0)  
      {
         // process if a buyer
          if (buyp[inv]>tradeprice)
          {
            // set feasible range from 0 to amount that would put current consumption below minimum allowed
               kleft = 0;
               kright =  (consumptions[consrow+0] - minconsumptions[consrow+0])/tradeprice;
            // compute excess of reservation prices over tradeprice at each point
               rpleft  = apgetrpk(inv,kleft) - tradeprice;
               rpright = apgetrpk(inv,kright) - tradeprice;
            // if rpright is positive, use kright (since rpleft is positive and reservation price decreases with amount purchased
               if (rpright>=0)
               {
                  #ifdef MEMCHECK
                     if (inv>tradevsize) aperrorhalt("Memory error: aptraded2 02");
                  #endif   
                  tradev[inv] = kright;
                  break;
               }
           // iterate until a trade amount is found at which the reservation price is less than the buy threshhold
              done = 0;
              while (done == 0)
              {
                // set kmid to midpoint of kleft and kright and compute excess of reservation price over tradeprice
                    kmid = (kleft + kright)/2.0;

                // compute excess of reservation price over tradeprice
                    rpmid = apgetrpk(inv,kmid) - tradeprice;
                    
                // change kleft or kright so that there is still a sign change
                    if (rpmid>=0) { kleft = kmid;  rpleft  = rpmid; }
                    if (rpmid<0)  { kright = kmid; rpright = rpmid; }
              
                // test for completion -- reservation price for kleft (>tradeprice) is below buy threshhold
                    if ((rpleft + tradeprice) < buythresh) done=1;
              }
           // set tradev
              #ifdef MEMCHECK
                  if (inv>tradevsize) aperrorhalt("Memory error: aptraded2 03");
              #endif   
              tradev[inv] = kright;
        }
  
      // process if a seller
          if (sellp[inv]<tradeprice)
          {
           // set feasible range
             kright = 0;
             kleft = -INFINITY;
             for (state=0;state<nstates;state++)
                if (conschg[state]>0)
                {
                   kval = (minconsumptions[consrow+state] - consumptions[consrow+state])/conschg[state];
                   if (kval>kleft) kleft = kval;
                }    
           // compute excess of reservation prices over tradeprice at each point
               rpleft  = apgetrpk(inv,kleft) - tradeprice;
               rpright = apgetrpk(inv,kright) - tradeprice;
               
           // if rpleft is negative, use kleft (since reservation price decreases (gets less negative) with amount held
               if (rpleft<=0)
               {
                  #ifdef MEMCHECK
                     if (inv>tradevsize) aperrorhalt("Memory error: aptraded2 03");
                  #endif   
                  tradev[inv] = kleft;
                  break;
               }       
           // iterate until a trade amount is found at which the reservation price is greater than the sell threshhold
              done = 0;
              while (done == 0)
              {
                 // set kmid to midpoint of kleft and kright and compute excess of reservation price over tradeprice
                     kmid = (kleft + kright)/2.0;
          
                 // compute excess of reservation price over tradeprice               
                    rpmid = apgetrpk(inv,kmid) - tradeprice;
                    
                 // change kleft or kright so that there is still a sign change
                    if (rpmid>=0) { kleft = kmid;  rpleft  = rpmid; }
                    if (rpmid<0)  { kright = kmid; rpright = rpmid; }
                    
                 // test for completion -- reservation price for kright is greater than the threshhold
                    if ((rpright + tradeprice) > sellthresh) done=1;
              }
           // set tradev
               #ifdef MEMCHECK
                  if (inv>tradevsize) aperrorhalt("Memory error: aptraded2 03");
               #endif   
               tradev[inv] = kleft;       
          }      
          
       } // end section to be skipped if no trade possible   
       
       // add to net amount and total demand and supply
          netamount = netamount + tradev[inv];
          if (tradev[inv] > 0) totaldemand = totaldemand + tradev[inv];
          if (tradev[inv] < 0) totalsupply = totalsupply - tradev[inv];
          
    //      printf("sec %d inv %d tradev %g netamount %g\n",tradesecurity,inv,tradev[inv],netamount);c=getchar();
    } 
 
  // calculate tradeamount as the smaller of demand and supply
      tradeamount = totaldemand;
      if (totalsupply < totaldemand) tradeamount = totalsupply;
      
  // return net amount
       return netamount;

}